program BearlibDemo;

uses
  bearlibfov,
  BeaRLibTerminal,
  SysUtils,
  Math,
  BearLibMap,
  MyToken;

// Map layers:
const
  kOpacityLayer = 0;
  kFOVLayer = 1;
  kColorLayer = 2;

type
  TLight = record
    X, Y, Radius: integer;
    Enabled: boolean;
    Color: TColor;
  end;
  PLight = ^TLight;

var
  map_id: integer = -1;
  player_sight_radius: integer = 15;
  lights: array[0..9] of TLight;
  player: PLight = @lights[0];
  advanced: boolean = False;

  function LoadMap: boolean;
  var
    ff: TextFile;
    line: string;
    map_lines: array of string;
    aX, aY, Width, Height: integer;
    c: char;
    light, aradius: integer;

  begin
    Result := False;
    AssignFile(ff, 'map.txt');
    Reset(ff);
    Width := 0;
    Height := 0;
    while not EOF(ff) do
    begin
      ReadLn(ff, line);
      if line = '' then
        break;
      SetLength(map_lines, Height + 1);
      map_lines[Height] := line;
      Inc(Height);
      Width := max(Width, length(line));
    end;
    if Height = 0 then
      exit;
    map_id := map_alloc(Width, Height, 'integer, flag, integer');
    map_clear(map_id);
    for ay := 0 to length(map_lines) - 1 do
    begin
      line := map_lines[ay];
      for ax := 0 to length(line) - 1 do
      begin
        c := line[ax + 1];
        map_set(map_id, ax, ay, kOpacityLayer, 0);
        if c = '#' then
          map_set(map_id, ax, ay, kOpacityLayer, 1)
        else if c in ['@', '1'..'9'] then
        begin
          if c = '@' then
            light := 0
          else
            light := Ord(c) - Ord('0');
          with lights[light] do
          begin
            X := ax;
            Y := ay;
            //Enabled := True;
            //Radius := 5;
            //Color := color_from_argb(255, 255, 0, 0);
          end;
        end;
      end;
    end;

    while not EOF(ff) do
    begin
      ReadLn(ff, line);
      if (line = '') or (Pos('//', line) > 0) then
        continue;
      light := StrToInt(TokenStrS(line, ':'));
      line := trim(line);
      aradius := StrToInt(TokenStrS(line, ' '));
      if (light < 0) or (light > 9) then
        continue;
      with lights[light] do
      begin
        Radius := aradius;
        Color := color_from_name(Line);
        Enabled := True;
      end;
    end;
    Result := True;
    CloseFile(ff);
  end;

  procedure LightFOVCallback(map_id, x, y: integer; opaque: pointer); cdecl;
  var
    light: PLight;
    dx, dy, dist2, factor: single;
    radius2: integer;
    acolor: TColor;
    r, g, b, cr, cg, cb: Integer;
  begin
    light := PLight(opaque);
    dx := light^.x - x;
    dy := light^.y - y;
    dist2 := dx * dx + dy * dy;
    radius2 := light^.radius * light^.radius;
    factor := 1 - dist2 / radius2;
    if factor < 0 then
      factor := 0;

    acolor := TColor(map_get(map_id, x, y, kColorLayer));
    r := byte(acolor shr 16);
    g := byte(acolor shr 8);
    b := byte(acolor);
    cr := byte(light^.color shr 16);
    cg := byte(light^.color shr 8);
    cb := byte(light^.color);
    Inc(r, trunc(cr * factor));
    Inc(g, trunc(cg * factor));
    Inc(b, trunc(cb * factor));
    map_set(map_id, x, y, kColorLayer, color_from_argb(255, min(r,255), min(g,255), min(b,255)));
  end;


  procedure CalcLight(light: PLight);
  begin
    fov_calc_cb(map_id, kOpacityLayer, light^.x, light^.y, light^.radius,
      @LightFOVCallback, light);
  end;


  function AverageWallColor(x, y: integer): TColor;
  var
    Width, Height, Count: integer;
    r, g, b: integer;
    ax, ay: integer;
    acolor: integer;
  begin
    Width := map_width(map_id);
    Height := map_height(map_id);
    Count := 0;
    r := 0;
    g := 0;
    b := 0;
    for ax := x - 1 to x + 1  do
      for ay := y - 1 to y + 1 do
      begin
        if not InRange(ax, 0, Width - 1) or not InRange(y, 0, Height - 1) then
          continue; // OOB
        if map_get(map_id, ax, ay, kFOVLayer) <> 1 then
          continue; // Must be in player's FOV
        if map_get(map_id, ax, ay, kOpacityLayer) > 0 then
          continue; // Must be a floor cell
        acolor := map_get(map_id, ax, ay, kColorLayer);
        Inc(r, byte(acolor shr 16));
        Inc(g, byte(acolor shr 8));
        Inc(b, byte(acolor shr 0));
        Inc(Count);
      end;

    if Count > 0 then
      Result := color_from_argb(255, min(255, r div Count), min(255, g div Count), min(255, b div Count))
    else
      Result := color_from_argb(255,0,0,0);
  end;


  procedure CalcWalls;
  var
    x, y: integer;
  begin
    for y := 0 to map_height(map_id) - 1 do
      for x := 0 to map_width(map_id) - 1 do
      begin
        if map_get(map_id, x, y, kFOVLayer) <> 1 then
          continue; // Must be in player's FOV
        if map_get(map_id, x, y, kOpacityLayer) <> 1 then
          continue; // Must be a wall
        map_set(map_id, x, y, kColorLayer, AverageWallColor(x, y));
      end;
  end;

  procedure DrawSimple;
  var
    x, y, c: integer;
  begin
    // Wall colors
    CalcWalls();

    // Cells
    for y := 0 to map_height(map_id) - 1 do
      for x := 0 to map_width(map_id) - 1 do
      begin
        if map_get(map_id, x, y, kFOVLayer) <= 0 then continue; // // Must be in player's FOV
        if map_get(map_id, x, y, kOpacityLayer) > 0 then
          c := $2588
        else
          c := $E000+32+4;//ord('.');
        terminal_color(map_get(map_id, x, y, kColorLayer));
        //terminal_color(color_from_name('gray'));
        terminal_put(x, y, c);
      end;
  end;

  procedure FillAmbient;
  var
    x, y: integer;
    ambient: TColor;
  begin
    ambient := color_from_name('darkest gray');
    for y := 0 to map_height(map_id) - 1 do
      for x := 0 to map_width(map_id) - 1 do
        map_set(map_id, x, y, kColorLayer, ambient);
  end;

  procedure MovePlayer(dx, dy: integer);
  var
    nx, ny: integer;
  begin
    nx := player^.x + dx;
    ny := player^.y + dy;

    if not InRange(nx, 0, map_width(map_id) - 1) or not
      InRange(ny, 0, map_height(map_id) - 1) then // OOB
      exit;

    if map_get(map_id, nx, ny, kOpacityLayer) > 0 then // Wall
      exit;

    player^.x := nx;
    player^.y := ny;
  end;

var
  key: integer;
  i: integer;
begin

  terminal_open;
  if not LoadMap then
  begin
    writeln('Couldn''t load a map from "map.txt" file');
    halt(-1);
  end;
  terminal_set(Format('size=%dx%d', [map_width(map_id), map_height(map_id)+1]));
  terminal_set('title=Lighting');
  terminal_set('font.name=media/UbuntuMono-R.ttf; font.size=12');
  //0xE000: tileset.png, size=16x16, spacing=2x1;
  terminal_set('0xE000: media/supplement.png, size=8x16, spacing=0x0;');

  while True do
  begin
    terminal_clear;
    FillAmbient;

    // Lights' and player's FOVs
    for i := 0 to 9 do
      if lights[i].Enabled then
        CalcLight(@lights[i]);
    map_clear_layer(map_id, kFOVLayer);
    fov_calc(map_id, kOpacityLayer, kFOVLayer, lights[0].x, lights[0].y,
      player_sight_radius);

    DrawSimple;

    for i := 0 to 9 do
      with lights[i] do
        if map_get(map_id, x, y, kFOVLayer) > 0 then
        begin
          if Enabled then
            terminal_color(Color)
          else
            terminal_color(color_from_name('gray'));
          if i = 0 then
            terminal_put(x, y, '@')
          else
            terminal_put(x, y, Ord('0') + i);
        end;

    terminal_color(color_from_name('white'));
    terminal_print(0, 24, 'Lighting mode [[use TAB]]: simple');
    terminal_refresh;
    key := terminal_read;
    case key of
      TK_CLOSE, TK_ESCAPE: break;
      TK_1..TK_9: lights[key - TK_1 + 1].Enabled := not lights[key - TK_1 + 1].Enabled;
      TK_LEFT: MovePlayer(-1, 0);
      TK_UP: MovePlayer(0, -1);
      TK_RIGHT: MovePlayer(+1, 0);
      TK_DOWN: MovePlayer(0, +1);
      TK_KP_1: MovePlayer(-1, +1);
      TK_KP_2: MovePlayer(0, +1);
      TK_KP_3: MovePlayer(+1, +1);
      TK_KP_4: MovePlayer(-1, 0);
      TK_KP_6: MovePlayer(+1, 0);
      TK_KP_7: MovePlayer(-1, -1);
      TK_KP_8: MovePlayer(0, -1);
      TK_KP_9: MovePlayer(+1, -1);
    end;

  end;
  terminal_close;
end.
